from .catboost_evaluation import *
from .evaluation_result import *
from .execution_case import *
from .utils import *

__all__ = [
    'EvalType',
    'CatboostEvaluation',
    'ScoreType',
    'ScoreConfig',
    'CaseEvaluationResult',
    'MetricEvaluationResult',
    'EvaluationResults',
    'calc_wilcoxon_test',
    'calc_bootstrap_ci_for_mean',
    'ExecutionCase',
    'make_dirs_if_not_exists',
    'series_to_line',
    'save_plot'
]



import json

from catboost import CatBoostError
from catboost.eval.factor_utils import FactorUtils
from catboost.core import _NumpyAwareEncoder


class ExecutionCase:

    def __init__(self,
                 params,
                 label=None,
                 ignored_features=None,
                 learning_rate=None):
        """
            Instances of this class are cases which will be compared during evaluation
            Params are CatBoost params
            label is a string which will be used for plots and other visualisations
            ignored_features is a set of additional feature indices to ignore
        """
        case_params = dict(params)

        if learning_rate is not None:
            case_params["learning_rate"] = learning_rate

        all_ignored_features = set()
        if "ignored_features" in case_params:
            all_ignored_features.update(set(case_params["ignored_features"]))
        if ignored_features is not None:
            all_ignored_features.update(ignored_features)

        case_params["ignored_features"] = list(all_ignored_features)

        self._label = label if label is not None else ""
        self._ignored_features = ignored_features
        self._ignored_features_str = FactorUtils.factors_to_ranges_string(self._ignored_features)

        self.__set_params(case_params)

    def __set_params(self, params):
        self._params = params
        self._params_hash = hash(json.dumps(self._params, sort_keys=True, cls=_NumpyAwareEncoder))

    def _set_thread_count(self, thread_count):
        if thread_count is not None and thread_count != -1:
            params = self._params
            params["thread_count"] = thread_count
            self.__set_params(params)

    @staticmethod
    def _validate_ignored_features(ignored_features, eval_features):
        for eval_feature in eval_features:
            if eval_feature in ignored_features:
                raise CatBoostError(
                    "Feature {} is in ignored set and in tmp-features set at the same time".format(eval_feature))

    def get_params(self):
        return dict(self._params)

    def get_label(self):
        return self._label

    def __str__(self):
        if len(self._label) == 0:
            return "Ignore: {}".format(self._ignored_features_str)
        else:
            return '{}'.format(self._label)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self._params == other._params and self._label == other._label

    def __hash__(self):
        return hash((self._label, self._params_hash))



"""
Module that contains different utils functions.
"""
import os


def make_dirs_if_not_exists(name):
    try:
        os.makedirs(name)
    except OSError:
        pass


def series_to_line(row, sep):
    return sep.join(map(str, row.tolist()))


def save_plot(fig, file_name=None):
    try:
        from plotly.offline import iplot, init_notebook_mode
        init_notebook_mode(connected=True)
        iplot(fig, filename=file_name)
    except:
        from plotly.offline import plot
        plot(fig, auto_open=False, filename=file_name)



from .core import (
    FeaturesData, EFstrType, EShapCalcType, EFeaturesSelectionAlgorithm, Pool, CatBoost,
    CatBoostClassifier, CatBoostRegressor, CatBoostRanker, CatBoostError, cv, train, sum_models, _have_equal_features,
    to_regressor, to_classifier, to_ranker, MultiRegressionCustomMetric, MultiRegressionCustomObjective
)  # noqa
from .version import VERSION as __version__  # noqa
__all__ = [
    'FeaturesData', 'EFstrType', 'EShapCalcType', 'EFeaturesSelectionAlgorithm', 'Pool', 'CatBoost',
    'CatBoostClassifier', 'CatBoostRegressor', 'CatBoostRanker', 'CatboostError',
    'cv', 'train', 'sum_models', '_have_equal_features',
    'to_regressor', 'to_classifier', 'to_ranker', 'MultiRegressionCustomMetric', 'MultiRegressionCustomObjective'
]

# API compatibility alias.
CatboostError = CatBoostError

try:
    from .widget import MetricVisualizer  # noqa
    __all__.append('MetricVisualizer')
except:
    pass



"""
Simple file reader. It intends to read lines from big files. Also it provides the group id of each line it reads.
"""

import pandas as pd
from .utils import series_to_line


class _BaseReader(object):
    def __init__(self, sep, group_feature_num):
        self._sep = sep
        self._group_feature_num = group_feature_num

    def get_separator(self):
        return self._sep

    def lines_generator(self):
        raise NotImplementedError("The base class don't have any lines_reader")

    def pack_lines_generator(self, pack_size):
        lines = []
        group_ids = []
        current_pack_size = 0

        lines_generator = self.lines_generator()
        for group_id, line in lines_generator:
            group_ids.append(group_id)
            lines.append(line)
            current_pack_size += 1
            if current_pack_size == pack_size:
                yield group_ids, lines
                lines = []
                group_ids = []
                current_pack_size = 0
        if current_pack_size != 0:
            yield group_ids, lines


class _SimpleStreamingFileReader(_BaseReader):
    def __init__(self, file_name, sep, has_header, group_feature_num=None):
        super(_SimpleStreamingFileReader, self).__init__(sep, group_feature_num)
        self._has_header = has_header
        self._file_name = file_name

    def lines_generator(self):
        with open(self._file_name, 'r') as file:
            if self._has_header:
                file.readline()
            for num, line in enumerate(file):
                if self._group_feature_num is None:
                    group_id = num
                else:
                    features = line.strip().split(self._sep, self._group_feature_num + 1)
                    group_id = features[self._group_feature_num]
                yield int(float(group_id)), line


# Can't handle big data. Can be used for tests.
class _SimpleDataReader(_BaseReader):
    def __init__(self, data, sep, group_feature_num=None):
        super(_SimpleDataReader, self).__init__(sep, group_feature_num)
        self._data = pd.DataFrame(data)

    def lines_generator(self):
        for num, (index, line) in enumerate(self._data.iterrows()):
            if self._group_feature_num is None:
                yield num, series_to_line(line, self._sep) + '\n'
            else:
                yield line.iloc[self._group_feature_num], series_to_line(line, self._sep) + '\n'

    def get_matrix(self):
        return self._data
